# --------------------------------------------------------------------
class DiceContext:
    def __init__(self,draw,darker=None):
        self._draw = draw
        self._darker = darker
        
        

    def __enter__(self):
        from wand.color import Color
        self._draw._draw.push()

        if self._darker:
            r, g, b = (self._draw._draw.fill_color.red_int8,
                       self._draw._draw.fill_color.green_int8,
                       self._draw._draw.fill_color.blue_int8)
            r *= self._darker
            g *= self._darker
            b *= self._darker
            #print(r,g,b)        
            self._draw._draw.fill_color = \
                Color(f'srgb({int(r)},{int(g)},{int(b)})')
        
        return self._draw

    def __exit__(self,*e):
        self._draw._draw.pop()

# --------------------------------------------------------------------
class DicePath:
    def __init__(self,draw):
        self._draw = draw

    def __enter__(self):
        self._draw._draw.path_start()
        return self

    def __exit__(self,*e):
        self._draw._draw.path_finish()

    def move(self,to):
        self._draw._draw.path_move(to=(self._draw.x(to[0]),
                                       self._draw.y(to[1])))
        
    def line(self,to):
        self._draw._draw.path_line(to=(self._draw.x(to[0]),
                                       self._draw.y(to[1])))

    def arc(self,to,r,cw=True):
        self._draw._draw.path_elliptic_arc(to=(self._draw.x(to[0]),
                                               self._draw.y(to[1])),
                                           radius=(self._draw.x(r),
                                                   self._draw.y(r)),
                                           clockwise=cw)
        
# --------------------------------------------------------------------
class DiceDraw:
    def __init__(self,width=100,height=100,fg='black',bg='white'):
        from wand.drawing import Drawing
        from wand.color import Color
        
        self._width  = width
        self._height = height
        self._fg     = fg if isinstance(fg,Color) else Color(fg)
        self._bg     = bg if isinstance(bg,Color) else Color(bg)
        self._draw   = Drawing()
        self._size   = min(self._width,self._height)

    @property
    def size(self):
        return self._size
    
    def x(self,xx):
        return int((xx  + 0.5) * self.size)

    def y(self,yy):
        return  int((0.5 -  yy) * self.size)

    def __enter__(self):
        self._draw.stroke_width = max(1,self.size//75)
        self._draw.stroke_color = self._fg
        self._draw.fill_color   = self._bg
        return self

    def number(self,num,yoff=0,scale=1):
        if num is None or num == '':
            return
        
        with DiceContext(self):
            self._draw.stroke_width   = 1
            self._draw.stroke_color   = self._fg
            self._draw.fill_color     = self._fg
            self._draw.text_alignment = 'center'
            self._draw.font_size      = scale * self.size / 2
            self._draw.text(self.x(0),
                            int(self.y(yoff)+self._draw.font_size//2),
                            f'{num}')
            
    def image(self):
        from wand.image import Image as WImage
        
        image = WImage(width=self._width,height=self._height,format='png')
        image.alpha_channel = True
        self._draw(image)
        
        off   = min(self._width,self._height) // 30
        copy  = image.clone()
        copy.shadow(50,off,0,0)
        copy.negate(channel='rgb')
        copy.composite(image,int(off/4),int(off/4))
        
        #copy.save(filename='d4.png')
        return copy

    def __exit__(self,*e):
        pass 

# --------------------------------------------------------------------
class DiceDrawer:
    def __init__(self,nsides,width,height,fg='red',bg='white'):
        from wand.color import Color
        self._nsides = nsides
        self._width  = width
        self._height = height
        self._fg     = Color(fg if isinstance(fg,str) else f'#{fg:06x}')
        self._bg     = Color(bg if isinstance(bg,str) else f'#{bg:06x}')
        if self._nsides not in [4,6,8,10,12,20]:
            raise RuntimeError(f'Unknown number of sides: {self._nsides}')
            

    def draw(self,num):
        if self._nsides ==  4:  return self.draw_d4 (num)
        if self._nsides ==  6:  return self.draw_d6 (num)
        if self._nsides ==  8:  return self.draw_d8 (num)
        if self._nsides == 10:  return self.draw_d10(num)
        if self._nsides == 12:  return self.draw_d12(num)
        if self._nsides == 20:  return self.draw_d20(num)
        return None
        
    def draw_d4(self,num):
        with DiceDraw(self._width,self._height,
                      fg=self._fg,bg=self._bg) as draw:
            with DicePath(draw) as path:
                path.move(to=( 0.000, 0.40))
                path.line(to=( 0.433,-0.35))
                path.line(to=(-0.433,-0.35))
                path.line(to=( 0.000, 0.40))

            draw.number(num)
            return draw.image()

    def draw_d6(self,num):
        with DiceDraw(self._width,self._height,
                      fg=self._fg,bg=self._bg) as draw:
            with DicePath(draw) as path:
                r = 0.05
                w = .4 - r
                path.move(to=(    w, 0.40))
                path.arc (to=( 0.40,    w), r=r)
                path.line(to=( 0.40,-   w))
                path.arc (to=(    w,-0.40), r=r)
                path.line(to=(-   w,-0.40))
                path.arc (to=(-0.40,-   w), r=r)
                path.line(to=(-0.40,    w))
                path.arc (to=(-   w, 0.40), r=r)
                path.line(to=(    w, 0.40))

            draw.number(num,yoff=.1)
            return draw.image()

    def draw_d8(self,num):
        with DiceDraw(self._width,self._height,
                      fg=self._fg,bg=self._bg) as draw:
            with DiceContext(draw,darker=.9):
                with DicePath(draw) as path:
                    path.move(to=(0.0000,0.5000))
                    path.line(to=(0.4330,0.2500))
                    path.line(to=(0.4330,-0.2500))
                    path.line(to=(0.0000,-0.5000))
                    path.line(to=(-0.4330,-0.2500))
                    path.line(to=(-0.4330,0.2500))
                    path.line(to=(0.0000,0.5000))
                        
            with DicePath(draw) as path:
                path.move(to=(0.0000,0.5000))
                path.line(to=(0.4330,-0.2500))
                path.line(to=(-0.4330,-0.2500))
                path.line(to=(0.0000,0.5000))

            draw.number(num,yoff=.1)
            return draw.image()

    def draw_d10(self,num):
        with DiceDraw(self._width,self._height,
                      fg=self._fg,bg=self._bg) as draw:
            with DiceContext(draw,darker=.9):
                with DicePath(draw) as path:
                    path.move(to=(0.0000,0.5000))            
                    path.line(to=(0.4750,0.1000))
                    path.line(to=(0.4750,-0.1000))
                    path.line(to=(0.0000,-0.5000))
                    path.line(to=(-0.4750,-0.1000))
                    path.line(to=(-0.4750,0.1000))
                    path.line(to=(0.0000,0.5000))
                    path.move(to=(0.2940,-0.1540))
                    path.line(to=(0.4750,-0.1000))
                    path.move(to=(-0.4750,-0.1000))
                    path.line(to=(-0.2940,-0.1540))
                    path.move(to=(0.0000,-0.5000))
                    path.line(to=(0.0000,-0.3000))
            with DicePath(draw) as path:
                path.move(to=(0.0000,0.5000))
                path.line(to=(0.2940,-0.1540))
                path.line(to=(0.0000,-0.3000))
                path.line(to=(-0.2940,-0.1540))
                path.line(to=(0.0000,0.5000))

            draw.number(num,yoff=.1)
            return draw.image()

    def draw_d12(self,num):
        with DiceDraw(self._width,self._height,
                      fg=self._fg,bg=self._bg) as draw:
            with DiceContext(draw,darker=.9):
                with DicePath(draw) as path:
                    path.move(to=( 0.0000, 0.5000))
                    path.line(to=( 0.2940, 0.4050))
                    path.line(to=( 0.4750, 0.1730))
                    path.line(to=( 0.4750,-0.1730))
                    path.line(to=( 0.2940,-0.4050))
                    path.line(to=( 0.0000,-0.5000))
                    path.line(to=(-0.2940,-0.4050))
                    path.line(to=(-0.4750,-0.1730))
                    path.line(to=(-0.4750, 0.1730))
                    path.line(to=(-0.2940, 0.4050))
                    path.line(to=( 0.0000, 0.5000))
                    path.line(to=( 0.0000, 0.3490))
                    path.move(to=( 0.4750, 0.1730))
                    path.line(to=( 0.3320, 0.1080))
                    path.move(to=( 0.2940,-0.4050))
                    path.line(to=( 0.2050,-0.2820))
                    path.move(to=(-0.2940,-0.4050))
                    path.line(to=(-0.2050,-0.2820))
                    path.move(to=(-0.4750, 0.1730))
                    path.line(to=(-0.3320, 0.1080))
            with DicePath(draw) as path:
                path.move(to=(0.0000,0.3490))
                path.line(to=(0.3320,0.1080))
                path.line(to=(0.2050,-0.2820))
                path.line(to=(-0.2050,-0.2820))
                path.line(to=(-0.3320,0.1080))
                path.line(to=(0.0000,0.3490))
                    

            draw.number(num,yoff=.1)
            return draw.image()
            
    def draw_d20(self,num):
        with DiceDraw(self._width,self._height,
                      fg=self._fg,bg=self._bg) as draw:
            with DiceContext(draw,darker=.85):
                with DicePath(draw) as path:
                    path.move(to=(0.0000,0.5000))
                    path.line(to=(0.4540,0.2620))
                    path.line(to=(0.4540,-0.2620))
                    path.line(to=(0.0000,-0.5000))
                    path.line(to=(-0.4540,-0.2620))
                    path.line(to=(-0.4540,0.2620))
                    path.line(to=(0.0000,0.5000))            
                    path.line(to=(0.0000,0.2920))
            with DiceContext(draw,darker=.95):
                with DicePath(draw) as path:
                    path.move(to=(0.0000,0.2920))
                    path.line(to=(0.4540,0.2620))
                    path.line(to=(0.2530,-0.1460))
                    path.line(to=(0.0000,-0.5000))
                    path.line(to=(-0.2530,-0.1460))
                    path.line(to=(-0.4540,0.2620))
                    path.line(to=(0.0000,0.2920))
                    path.move(to=(0.4540,-0.2620))
                    path.line(to=(0.2530,-0.1460))
                    path.move(to=(-0.4540,-0.2620))
                    path.line(to=(-0.2530,-0.1460))
            with DicePath(draw) as path:
                path.move(to=(0.0000,0.2920))
                path.line(to=(0.2530,-0.1460))
                path.line(to=(-0.2530,-0.1460))
                path.line(to=(0.0000,0.2920))
            

            scale = .7
            yoff  = .07
            if num > 9:
                scale *= .8
                yoff  =  .015
            draw.number(num,yoff=yoff,scale=scale)
            return draw.image()


# --------------------------------------------------------------------
class DiceAnimator:
    def __init__(self,nsides,width,height,fg='red',bg='white'):
        drawer = DiceDrawer(nsides = nsides,
                            width  = width,
                            height = height,
                            fg     = fg,
                            bg     = bg)
        self._pool = [drawer.draw(v) for v in range(1,nsides+1)]
            
    def draw(self,num,n,dt=10):
        
        '''
        delay : int
            1 / 100 of a second as base delay
        '''
        from random     import sample
        from wand.image import Image

        if num > len(self._pool):
            raise RuntimeError(f'Final value {num} not possible for '
                               f'd{len(self._pool)}')

        if n >= len(self._pool) - 1:
            raise RuntimeError(f'Pre-steps {n} not possible for '
                               f'd{len(self._pool)}')
            
        end   = self._pool[num-1]

        # Round-about, but for debug 
        nums  = list(range(1,len(self._pool)+1))
        nums.remove(num)
        pool  = sample(nums, k=n)
        other = [self._pool[i-1] for i in pool]
        print(num,pool)
        
        # More direct
        # pool  = self._pool[:num] + self._pool[num+1:]
        # other = sample(pool, k=n)
        
        faces = other + [end]
        dang  = 360 / (len(faces))
        img   = Image(format='gif')
        ang   = 360
        wait  = dt
        for i, face in enumerate(faces):
            ang  -= dang
            wait += dt

            with face.clone() as rotated:
                w, h = rotated.size
                rotated.rotate(ang, reset_coords=False)
                rotated.crop(0,0,w,h)
                rotated.dispose = 'previous'
                rotated.delay   = wait # dt * (i + 1)
                rotated.loop    = 1
                img.sequence.append(rotated)

        # Make sure we dispose of the previous image altogether
        #img.dispose = 'previous'
        img.coalesce()
        #img.deconstruct()
        

        return img

#
# EOF
#



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
## END_IMPORT

# --------------------------------------------------------------------
class DeselectTrait(Trait):
    ID = 'deselect'
    THIS = 'D' # Deselect only this piece
    ALL  = 'A' # Deselect all pieces 
    ONLY = 'S' # Select this piece only
    def __init__(self,
                 command     = '',
                 key         = '',
                 description = '',
                 unstack     = False,
                 deselect    = THIS):
        '''Create a deselect trait'''
        super(DeselectTrait,self).__init__()
        self.setType(command     = command,
                     key         = key,
                     description = description,
                     unstack     = unstack,
                     deselect    = deselect)
        self.setState()


Trait.known_traits.append(DeselectTrait)

#
# EOF
#



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
## END_IMPORT

class CargoTrait(Trait):
    ID = 'matPiece'
    NO_MAT = 'noMat'

    def __init__(self,
                 description = '',
                 maintainRelativeFacing = True,
                 detectionDistanceX     = 0,
                 detectionDistanceY     = 0,
                 attachKey              = '',
                 detachKey              = ''):
        '''Create cargo trait

        '''
        self.setType(description            = description,
                     maintainRelativeFacing = maintainRelativeFacing,
                     detectionDistanceX     = detectionDistanceX,
                     detectionDistanceY     = detectionDistanceY,
                     attachKey              = attachKey,
                     detachKey              = detachKey)
        self.setState(mat = CargoTrait.NO_MAT)

Trait.known_traits.append(CargoTrait)
        
#
# EOF
#



#
from . base         import *
from . dicehand     import *
from . piece        import *
from . map          import *
from . scenario     import *
from . countersheet import *
from . gamebox      import *
from . exporter     import *
from . ztexp        import *

#
# EOF
#



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
## END_IMPORT

# --------------------------------------------------------------------
class CalculatedTrait(Trait):
    ID = 'calcProp'
    def __init__(self,name='',expression='',description=''):
        '''Define a trait that calculates a property'''
        super(CalculatedTrait,self).__init__()
        self.setType(name        = name,
                     expression  = expression,
                     description = description)
        self.setState()


Trait.known_traits.append(CalculatedTrait)

#
# EOF
#



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
## END_IMPORT


class MatTrait(Trait):
    ID = 'mat'

    def __init__(self,
                 name = 'Mat',
                 description = ''):
        '''Create mat trait

        '''
        self.setType(name        = name,
                     description = description)
        self.setState(content='0')

    def setContent(self,*args):
        # Not sure this is correct 
        self.setState(content=str(len(args))+';'+';'.joint(args))


Trait.known_traits.append(MatTrait)
	        
#
# EOF
#



## BEGIN_IMPORT
from .. common import VerboseGuard
from . base import *
from . element import Element
## END_IMPORT

# --------------------------------------------------------------------
class TurnLevel(Element):
    UNIQUE = ['property']
    def __init__(self,elem,tag,node=None,**kwargs):
        super(TurnLevel,self).__init__(elem,tag,node=node,**kwargs)

    def addLevel(self,counter=None,phases=None):
        '''Add a `Level` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Level
            The added element
        '''
        if counter is None and phases is None:
            return self
        
        t = TurnCounter if counter is not None else TurnList
        o = counter     if counter is not None else phases

        subcounter = o.pop('counter',None)
        subphases  = o.pop('phases',None)

        s = t(self,node=None,**o)

        return s.addLevel(subcounter, subphases)

    def getUp(self):
        return self.getParent(TurnLevel)
    def addCounter(self,**kwargs):
        '''Add a `Counter` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Counter
            The added element
        '''
        return self.add(self,TurnCounter,**kwargs)
    def addList(self,**kwargs):
        '''Add a `List` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : List
            The added element
        '''
        return self.add(self,TurnList,**kwargs)
    def getCounter(self):
        return self.getAllElements(TurnCounter)
    def getList(self):
        return self.getAllElements(TurnList)

# --------------------------------------------------------------------
class TurnTrack(TurnLevel):
    TAG = Element.MODULE+'turn.TurnTracker'
    UNIQUE   = ['name']
    MAXIMUM  = 'Maximum'
    FIXED    = 'Fixed'
    VARIABLE = 'Variable'
    def __init__(self,elem,node=None,
                 name             = '',
                 buttonText       = 'Turn',
                 hotkey           = '',
                 icon             = '',
                 length           = -1,
                 lengthStyle      = MAXIMUM,
                 nexthotkey       = key('T',ALT),
                 plusButtonSize   = 22,
                 prevhotkey       = key('T',ALT_SHIFT),
                 reportFormat     = 'Turn updated from $oldTurn$ to $newTurn$',
                 turnButtonHeight = 22,
                 fwdOnly          = True,
                 turnFormat       = None,
                 counter          = None,
                 phases           = None):
        levels = (counter if counter is not None else
                  phases if phases is not None else None)
        if levels is not None:
            lvl = 1
            lvls = [f'$level{lvl}$']
            sub  = levels
            while True:
                sub = sub.get('counter',sub.get('phases',None))
                if sub is None:
                    break
                lvl += 1
                lvls.append(f'$level{lvl}$')
            
            turnFormat = ' '.join(lvls)
        
        if turnFormat is None:
            turnFormat = '$level1$ $level2$ $level3$ $level4$'        
        
        super(TurnTrack,self).__init__(elem, self.TAG,
                                       node             = node,
                                       name             = name,
                                       buttonText       = buttonText,
                                       hotkey           = hotkey,
                                       icon             = icon,
                                       length           = length,
                                       lengthStyle      = lengthStyle,
                                       nexthotkey       = nexthotkey,
                                       plusButtonSize   = plusButtonSize,
                                       prevhotkey       = prevhotkey,
                                       reportFormat     = reportFormat,
                                       turnButtonHeight = turnButtonHeight,
                                       turnFormat       = turnFormat)

        self.addLevel(counter=counter, phases=phases)

    def getGame(self):
        return self.getParent(Game)
    def getLists(self,asdict=True):
        '''Get all List element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `List`
            elements.  If `False`, return a list of all List`
            children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `List` children

        '''
        return self.getElementsByKey(TurnList,'property',asdict=asdict)
    def getCounters(self,asdict=True):
        '''Get all Counter element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `Counter`
            elements.  If `False`, return a list of all Counter`
            children.

        Returns
        -------
        children : dict or list
            Dictionary or list of `Counter` children

        '''
        return self.getElementsByKey(TurnCounter,'property',asdict=asdict)
    def addHotkey(self,**kwargs):
        '''Add a `Hotkey` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Hotkey
            The added element
        '''
        return self.add(TurnGlobalHotkey,**kwargs)
    def getHotkeys(self,asdict=True):
        return self.getElementsByKey(TurnGlobalHotkey,'name',asdict=asdict)
    def encode(self):
        ret = f'TURN{self["name"]}\t'
        
        return []

registerElement(TurnTrack)

# --------------------------------------------------------------------
class TurnCounter(TurnLevel):
    TAG = Element.MODULE+"turn.CounterTurnLevel"
    def __init__(self,elem,node=None,
                 property      = '',
                 start         = 1,
                 incr          = 1,
                 loop          = False,
                 loopLimit     = -1,
                 turnFormat    = "$value$"):
        super(TurnCounter,self).__init__(elem,self.TAG,node=node,
                                         property       = property,
                                         start          = start,
                                         incr           = incr,
                                         loop           = loop,
                                         loopLimit      = loopLimit,
                                         turnFormat     = turnFormat)
                    
registerElement(TurnCounter)

# --------------------------------------------------------------------
class TurnList(TurnLevel):
    TAG = Element.MODULE+"turn.ListTurnLevel"
    def __init__(self,elem,node=None,
                 property      = '',
                 names         = [],
                 configFirst   = False,
                 configList    = False,
                 turnFormat    = '$value$'):
        super(TurnList,self).\
            __init__(elem,self.TAG,node=node,
                     property       = property,
                     list           = ','.join([str(p) for p in names]),
                     configFirst    = configFirst,
                     configList     = configList,
                     turnFormat     = turnFormat)
                  
registerElement(TurnList)

# --------------------------------------------------------------------
class TurnGlobalHotkey(Element):
    TAG = Element.MODULE+'turn.TurnGlobalHotkey'
    UNIQUE = ['name']
    def __init__(self,elem,
                 node         = None,
                 hotkey       = '',
                 match        = '{true}',
                 reportFormat = '',
                 name         = ''):
        '''Global key activated by turn change

        Parameters
        ----------
        doc : Element
            Parent
        node : xml.dom.Element
            Node to read state from
        hotkey : str
            What to send (global command)
        match : str
            When to send
        reportFormat : str
            What to what
        name : str
            A free form name
        '''
        super(TurnGlobalHotkey,self).__init__(elem,self.TAG,
                                              node         = node,
                                              hotkey       = hotkey,
                                              match        = match,
                                              reportFormat = reportFormat,
                                              name         = name)

    def getTurnTrack(self):
        '''Get the turn track'''
        return self.getParent(TurnTrack)

registerElement(TurnGlobalHotkey)

#
# EOF
#



# --------------------------------------------------------------------
class ZTPiece:
    def __init__(self,front_image,back_image,x,y,terrain,card):
        self._front   = front_image
        self._back    = back_image
        self._terrain = terrain
        self._card    = card
        self._x       = x
        self._y       = y

    @property
    def two_sides(self):
        return (self._front        is not None and
                self._back         is not None)

    @property
    def size(self):
        return self.width,self.height

    @property
    def width(self):
        return self._front.width

    @property
    def height(self):
        return self._front.height

    
    def __str__(self):
        return (f'  Piece: '+
                f'{"card" if self._card else "piece"} '+
                f'{"terrain" if self._terrain else "piece"} '+
                f'front={self._front} back={self._back}')

#
# EOF
#



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
from .. base import *
# from .. widget import *
# from .. withtraits import PieceSlot
## END_IMPORT

# --------------------------------------------------------------------
class PlaceTrait(Trait):
    ID      = 'placemark'
    STACK_TOP = 0
    STACK_BOTTOM = 1
    ABOVE = 2
    BELOW = 3

    # How the LaTeX exporter organises the units.  Format with
    # 0: the group
    # 1: the piece name 
    # SKEL_PATH = (PieceWindow.TAG +r':Counters\/'        +
    #              TabWidget.TAG   +r':Counters\/'        +
    #              PanelWidget.TAG +':{0}'         +r'\/'+
    #              ListWidget.TAG  +':{0} counters'+r'\/'+
    #              PieceSlot.TAG   +':{1}')
    @classmethod
    # @property
    def SKEL_PATH(cls):
## BEGIN_IMPORT
        from .. widget import PieceWindow, TabWidget, PanelWidget, ListWidget
        from .. withtraits import PieceSlot
## END_IMPORT

        return (PieceWindow.TAG +r':Counters\/'        +
                TabWidget.TAG   +r':Counters\/'        +
                PanelWidget.TAG +':{0}'         +r'\/'+
                ListWidget.TAG  +':{0} counters'+r'\/'+
                PieceSlot.TAG   +':{1}')
    
    def __init__(self,
                 command         = '', # Context menu name
                 key             = '', # Context menu key
                 markerSpec      = '', # Full path in module
                 markerText      = 'null', # Hard coded message
                 xOffset         = 0,
                 yOffset         = 0,
                 matchRotation   = True,
                 afterKey        = '',
                 description     = '',
                 gpid            = '', # Set in JAVA, but with warning
                 placement       = ABOVE,
                 above           = False):
        '''Create a place marker trait (VASSAL.counter.PlaceMarker)'''
        super(PlaceTrait,self).__init__()
        self.setType(command         = command,          # Context menu name
                     key             = key,              # Context menu key
                     markerSpec      = markerSpec,
                     markerText      = markerText,
                     xOffset         = xOffset,
                     yOffset         = yOffset,
                     matchRotation   = matchRotation,
                     afterKey        = afterKey,
                     description     = description,
                     gpid            = gpid,
                     placement       = placement,
                     above           = above)
        self.setState()

Trait.known_traits.append(PlaceTrait)

# --------------------------------------------------------------------
class ReplaceTrait(PlaceTrait):
    ID = 'replace'
    def __init__(self,
                 command         = '', # Context menu name
                 key             = '', # Context menu key
                 markerSpec      = '', # Full path in module
                 markerText      = 'null', # Hard message
                 xOffset         = 0,
                 yOffset         = 0,
                 matchRotation   = True,
                 afterKey        = '',
                 description     = '',
                 gpid            = '', # Set in JAVA
                 placement       = PlaceTrait.ABOVE,
                 above           = False):
        super(ReplaceTrait,self).__init__(command         = command, 
                                          key             = key,  
                                          markerSpec      = markerSpec,
                                          markerText      = markerText,
                                          xOffset         = xOffset,
                                          yOffset         = yOffset,
                                          matchRotation   = matchRotation,
                                          afterKey        = afterKey,
                                          description     = description,
                                          gpid            = gpid,
                                          placement       = placement,
                                          above           = above)
    

Trait.known_traits.append(ReplaceTrait)

#
# EOF
#



## BEGIN_IMPORT
from .. common import VerboseGuard
from . base import *
from . element import Element
from . gameelements import GameElement
## END_IMPORT

# ====================================================================
def createKeyHelp(*args,**kwargs):
    '''Creates a help file with key-bindings

    See Documentation.createKeyHelp
    '''
    return Documentation.createKeyHelp(*args,**kwargs)

# --------------------------------------------------------------------
class Documentation(GameElement):
    TAG=Element.MODULE+'Documentation'
    def __init__(self,doc,node=None,**kwargs):
        '''Documentation (or help menu)

        Parameters
        ----------
        doc : Element
            Parent
        node : xml.dom.Element
            Node to read state from
        kwargs : dict
            Attributes
        '''
        super(Documentation,self).__init__(doc,self.TAG,node=node,**kwargs)

    def addAboutScreen(self,**kwargs):
        '''Add a `AboutScreen` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : AboutScreen
            The added element
        '''
        return self.add(AboutScreen,**kwargs)
    def addHelpFile(self,**kwargs):
        '''Add a `HelpFile` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : HelpFile
            The added element
        '''
        return self.add(HelpFile,**kwargs)
    def addBrowserHelpFile(self,**kwargs):
        '''Add a `BrowserHelpFile` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : BrowserHelpFile
            The added element
        '''
        return self.add(BrowserHelpFile,**kwargs)
    def addBrowserPDFFile(self,**kwargs):
        '''Add a `BrowserPDFFile` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : BrowserPDFFile
            The added element
        '''
        return self.add(BrowserPDFFile,**kwargs)
    def addTutorial(self,**kwargs):
        '''Add a `Tutorial` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Tutorial
            The added element
        '''
        return self.add(Tutorial,**kwargs)
    def getAboutScreens(self):
        return self.getElementsByKey(AboutScreen,'title')
    def getHelpFiles(self):
        return self.getElementsByKey(HelpFile,'title')
    def getBrowserHelpFiles(self):
        return self.getElementsByKey(BrowserHelpFile,'title')
    def getBrowserPDFFiles(self):
        return self.getElementsByKey(BrowserPDFFile,'title')
    def getTutorials(self):
        return self.getElementsByKey(Tutorial,'name')

    @classmethod
    def createKeyHelp(cls,keys,title='',version=''):
        '''Creates a help file with key-bindings
        
        Parameters
        ----------
        keys : list of list of str
             List of key-binding documentation
        title : str
             Title of help file
        version : str
             Version number
        
        Returns
        -------
        txt : str
            File content
        '''
        txt = f'''
        <html>
         <body>
          <h1>{title} (Version {version}) Key bindings</h1>
          <table>
          <tr><th>Key</th><th>Where</th><th>Effect</th></tr>'''
        
        for key, where, description in keys:
            txt += (f'<tr><td>{key}</td>'
                    f'<td>{where}</td>'
                    f'<td>{description}</td></tr>')
        
        txt += '''
          </table>
         </body>
        </html>'''
        
        return txt 

registerElement(Documentation)

# --------------------------------------------------------------------
class AboutScreen(Element):
    TAG = Element.MODULE+'documentation.AboutScreen'
    UNIQUE = ['title']
    def __init__(self,doc,node=None,title='',fileName=""):
        '''Create an about screen element that shows image

        Parameters
        ----------
        doc : Element
            Parent
        node : xml.dom.Element
            Node to read state from
        title : str
            Entry title
        fileName : str
            Internal file name        
        '''
        super(AboutScreen,self).__init__(doc,
                                         self.TAG,
                                         node     = node,
                                         fileName = fileName,
                                         title    = title)
    def getDocumentation(self):
        '''Get Parent element'''
        return self.getParent(Documentation)

registerElement(AboutScreen)

# --------------------------------------------------------------------
class BrowserPDFFile(Element):
    TAG = Element.MODULE+'documentation.BrowserPDFFile'
    UNIQUE = ['title','pdfFile']
    def __init__(self,doc,node=None,title='',pdfFile=''):
        '''Create help menu item that opens an embedded PDF
        
        Parameters
        ----------
        doc : Element
            Parent
        node : xml.dom.Element
            Node to read state from
        title : str
            Entry title
        pdfFile : str
            Internal file name
        '''
        super(BrowserPDFFile,self).__init__(doc,self.TAG,
                                            node    = node,
                                            pdfFile = pdfFile,
                                            title   = title)
    def getDocumentation(self):
        '''Get Parent element'''
        return self.getParent(Documentation)

registerElement(BrowserPDFFile)
    
# --------------------------------------------------------------------
class HelpFile(Element):
    TAG = Element.MODULE+'documentation.HelpFile'
    ARCHIVE = 'archive'
    UNIQUE = ['title','fileName']
    def __init__(self,doc,node=None,
                 title='',
                 fileName='',
                 fileType=ARCHIVE):
        '''Create a help menu entry that opens an embeddded file
        
        Parameters
        ----------
        doc : Element
            Parent
        node : xml.dom.Element
            Node to read state from
        title : str
            Entry title
        fileName : str
            Internal file name
        fileType : str
            How to find the file 
        '''
        super(HelpFile,self).__init__(doc,self.TAG,node=node,
                                      fileName = fileName,
                                      fileType = fileType,
                                      title    = title)

    def getDocumentation(self):
        '''Get Parent element'''
        return self.getParent(Documentation)

registerElement(HelpFile)
    
# --------------------------------------------------------------------
class BrowserHelpFile(Element):
    TAG = Element.MODULE+'documentation.BrowserHelpFile'
    UNIQUE = ['title','startingPage']
    def __init__(self,doc,node=None,
                 title='',
                 startingPage='index.html'):
        '''Create a help menu entry that opens an embeddded HTML
        page (with possible sub-pages) file
        
        Parameters
        ----------
        doc : Element
            Parent
        node : xml.dom.Element
            Node to read state from
        title : str
            Entry title
        startingPage : str
            which file to start at
        '''
        super(BrowserHelpFile,self).__init__(doc,self.TAG,node=node,
                                             startingPage=startingPage,
                                             title=title)

    def getDocumentation(self):
        '''Get Parent element'''
        return self.getParent(Documentation)

registerElement(BrowserHelpFile)
    
# --------------------------------------------------------------------
class Tutorial(Element):
    TAG = Element.MODULE+'documentation.Tutorial'
    UNIQUE = ['name','logfile']
    def __init__(self,doc,node=None,
                 name            = 'Tutorial',
                 logfile         = 'tutorial.vlog',
                 promptMessage   = 'Load the tutorial?',
                 welcomeMessage  = 'Press "Step forward" (PnDn) to step through the tutorial',
                 launchOnStartup = True):
        '''Add a help menu item that loads the tutorial

        Also adds the start-up option to run the tutorial


        Parameters
        ----------
        doc : Element
            Parent
        node : xml.dom.Element
            Node to read state from
        name : str
            Name of entry
        logfile : str
            Internal file name
        promptMessage : str
            What to show
        launchOnStartup : bool
            By default, launch tutorial first time running module
        '''
        super(Tutorial,self).__init__(doc,self.TAG,node=node,
                                      name            = name,
                                      logfile         = logfile,
                                      promptMessage   = promptMessage,
                                      welcomeMessage  = welcomeMessage,
                                      launchOnStartup = launchOnStartup)

    def getDocumentation(self):
        '''Get Parent element'''
        return self.getParent(Documentation)

registerElement(Tutorial)
    

#
# EOF
#



## BEGIN_IMPORTS
from pywargame.common import VerboseGuard
from . base import DomInspect
## END_IMPORTS

# --------------------------------------------------------------------        
class ZTStack:
    def __init__(self,x,y,ids,flp,rot):
        self._x    = x
        self._y    = y
        self._ids  = ids
        self._flip = flp
        self._rot  = rot

    def translate(self,dx,dy):
        self._x += dx
        self._y += dy

    def scale(self,f):
        self._x *= f
        self._y *= f
    
        
# --------------------------------------------------------------------        
class ZTLayout(DomInspect):
    def __init__(self,elem):
        self._name    = self._get_attr(elem,'board')
        self._left    = float(self._get_attr(elem,'left',  -1))
        self._top     = float(self._get_attr(elem,'top',   -1))
        self._right   = float(self._get_attr(elem,'right', -1))
        self._bottom  = float(self._get_attr(elem,'bottom',-1))
        self._visible = self._get_attr(elem,'visible','true')=='true'

        self._stacks  = [self.parse_stack(s)
                         for s in self._find_children(elem,'counter','stack')]

    @property
    def ulx(self):
        return self._left
    
    @property
    def lrx(self):
        return self._right
    
    @property
    def uly(self):
        return self._top
    
    @property
    def lry(self):
        return self._bottom

    @property
    def left(self):
        return int(self._left)

    @property
    def top(self):
        return int(self._top)
    
    @property
    def width(self):
        return int(self._right - self._left+.5)

    @property
    def height(self):
        return int(self._bottom - self._top+.5)
    
    def parse_stack(self,elem):
        x   = float(self._get_attr(elem,'x',0))
        y   = float(self._get_attr(elem,'y',0))
        # We ignore the `rot` and `Side` attributes for now
        # More difficult to propagate to the states, which are handled
        # in VSav.  Could perhaps add a mapping from traits to states
        # when the traits are flattened there. 
        ids = []
        flp = []
        rot = [] 
        if elem.tagName == 'counter':
            ids.append(int(self._get_attr(elem,'id',-1)))
            flp.append(self._get_attr(elem,'side','')=='Back')
            rot.append(int(self._get_attr(elem,'rot',0)))
        else:
            ids.extend([int(self._get_attr(c,'id',-1))
                        for c in self._find_children(elem,'counter')])
            flp.extend([self._get_attr(c,'side','')=='Back'
                        for c in self._find_children(elem,'counter')])
            rot.extend([int(self._get_attr(c,'rot',0))
                        for c in self._find_children(elem,'counter')])

        return ZTStack(x,y,ids,flp,rot)
            
    def translate(self,dx,dy):
        self._left   += dx
        self._right  += dx
        self._top    += dy
        self._bottom += dy
        for s in self._stacks:
            s.translate(dx,dy)

    def scale(self,f):
        self._left   *= f
        self._right  *= f
        self._top    *= f
        self._bottom *= f
        for s in self._stacks:
            s.scale(f)

    def __str__(self):
        return f'Layout "{self._name}" ({len(self._stacks)}: {self.ulx},{self.uly},{self.lrx},{self.lry}'
            
# --------------------------------------------------------------------        
class ZTScenario(DomInspect):
    def __init__(self,file):        
        from xml.dom.minidom import parse
        from pathlib import Path
        
        with VerboseGuard(f'Got a scenario {file}') as v:
            self._dom = parse(file)

            p = Path(file if isinstance(file,str) else file.name)

            self._file = p.stem
            self._root = self._dom.firstChild
            self._game = self._get_attr(self._root,'game-box')
            self._name = self._get_attr(self._root,'scenario-name')
            self._desc = self._get_attr(self._root,'scenario-description','')
            self._copy = self._get_attr(self._root,'scenario-copyright','')

            self._layouts = [self.parse_layout(elem)
                             for elem in
                             self._root.getElementsByTagName('layout')]

            self.ulx = None
            self.uly = None
            self.lrx = None
            self.lry = None
            
            v(f'Bounding box: {self.bounding_box}')

    def parse_layout(self,elem):
        return ZTLayout(elem)

    def translate(self,dx,dy):
        with VerboseGuard(f'Translating scenario by {dx},{dy}'):
            self.ulx = None
            self.uly = None
            self.lrx = None
            self.lry = None
            
            for l in self._layouts:
                l.translate(dx,dy)

    def scale(self,f):
        with VerboseGuard(f'Scaling scenario by {f}'):
            self.ulx = None
            self.uly = None
            self.lrx = None
            self.lry = None
            
            for l in self._layouts:
                l.scale(f)
                
    def on_map(self,name):
        return any([l._name == name for l in self._layouts])

    def map_layout(self,name):
        return [l for l in self._layouts if l._name == name]

    def calc_bounding_box(self,name=None):
        layouts = self._layouts if name is None else self.map_layout(name)
        ulx     = int(min([l.ulx for l in layouts])-.5)
        uly     = int(min([l.uly for l in layouts])-.5)
        lrx     = int(max([l.lrx for l in layouts])+.5)
        lry     = int(max([l.lry for l in layouts])+.5)
        return ulx,uly,lrx,lry
        
    @property 
    def bounding_box(self):
        '''Cache calculations'''
        if self.ulx is None or \
           self.uly is None or \
           self.lrx is None or \
           self.lry is None:
            self.ulx = int(min([l.ulx for l in self._layouts])-.5)
            self.uly = int(min([l.uly for l in self._layouts])-.5)
            self.lrx = int(max([l.lrx for l in self._layouts])+.5)
            self.lry = int(max([l.lry for l in self._layouts])+.5)

        return self.ulx,self.uly,self.lrx,self.lry

    def __str__(self):
        return f'Scenario "{self._name}": {self._desc}'+'\n'+\
            '\n'.join([str(l) for l in self._layouts])
        
    @property
    def width(self):
        x1,_,x2,_ = self.bounding_box
        return x2-x1 
    @property
    def height(self):
        _,y1,_,y2 = self.bounding_box
        return y2-y1
   
    
#
# EOF
#



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
## END_IMPORT

# --------------------------------------------------------------------
class GlobalHotkeyTrait(Trait):
    ID = 'globalhotkey'
    def __init__(self,
                 name          = '', # Command received
                 key           = '', # Command key received
                 globalHotkey  = '', # Key to send
                 description   = ''):
        '''Create a global key command in piece
        (VASSAL.counters.GlobalHotkey)'''
        self.setType(name          = name,
                     key           = key,
                     globalHotkey  = globalHotkey,
                     description   = description)
        self.setState()
        
Trait.known_traits.append(GlobalHotkeyTrait)

#
# EOF
#



## BEGIN_IMPORT
from .. common import VerboseGuard
from . image import GBXImage
from . base import CbFont
from . features import Features
## END_IMPORT

# ====================================================================
class GBXDraw:
    def __init__(self,ar):
        '''Base class for drawing objects'''
        self._flags  = ar.dword()
        self._left   = ar.word()
        self._top    = ar.word()
        self._right  = ar.word()
        self._bottom = ar.word()

    def isSecondPass(self):
        return self._flags & 0x00000008

    def bbWidth(self):
        return self._right - self._left

    def bbHeight(self):
        return self._bottom - self._top

    def centerX(self):
        return (self._left + self._right)//2

    def centerY(self):
        return (self._top + self._bottom)//2

    def center(self):
        return (self.centerX(),self.centerY())

    def upperLeft(self):
        return (self._left,self._top)

    def lowerRight(self):
        return (self._right,self._bottom)

    def bbSize(self):
        return (self.bbWidth(),self.bbHeight())

    @classmethod
    def hex(cls,val):
        if val == 0xFF000000:
            h = 'none'
        else:
            b = (val >> 16) & 0xFF
            g = (val >>  8) & 0xFF
            r = (val >>  0) & 0xFF
            h = f'rgb({r},{g},{b})'
        return h

    def toDict(self,calc):
        return None

    def baseDict(self):
        return {
            'left':   self._left,
            'top':    self._top,
            'right':  self._right,
            'bottom': self._bottom,
            'x':      self.centerX(),
            'y':      self.centerY()
        }
    
    def svg(self,dwg,g,defmap):
        print(f'{self.__class__}.svg method not implemented')
        pass
    
    def __str__(self):
        return (f'Flags:{self._flags:08x} '
                + f'({self._left},{self._top})x({self._right},{self._bottom})')

# --------------------------------------------------------------------
class GBXRectangle(GBXDraw):
    def __init__(self,ar):
        '''Draw a rectangle'''
        with VerboseGuard(f'Reading rectangle'):
            super(GBXRectangle,self).__init__(ar)
            self._fill  = ar.dword()
            self._line  = ar.dword()
            self._width = ar.word()

    def svg(self,dwg,g,defmap):
        r = g.add(dwg.rect(insert=(self.upperLeft()),
                           size=(self.bbSize()),
                           fill=self.hex(self._fill),
                           stroke=self.hex(self._line),
                           stroke_width=self._width))
        
    def __str__(self):
        return 'Rectangle: '+super(GBXRectangle,self).__str__()

# --------------------------------------------------------------------
class GBXEllipse(GBXRectangle):
    def __init__(self,ar):
        '''Draw an ellipse'''
        with VerboseGuard(f'Reading ellipse'):
            super(GBXEllipse,self).__init__(ar)

    def svg(self,dwg,g,defmap):
        '''Create SVG object'''
        g.add(dwg.ellipse(center=(self.centerX(),self.centerY()),
                          r=(self.bbWidth(),self.bbHeight()),
                          fill=self.hex(self._fill),
                          stroke=self.hex(self._line),
                          stroke_width=self._width))
    def __str__(self):
        return 'Ellipse: '+super(GBXRectangle,self).__str__()
        
# --------------------------------------------------------------------
class GBXLine(GBXDraw):
    def __init__(self,ar):
        '''Draw a line'''
        with VerboseGuard(f'Reading line'):
            super(GBXLine,self).__init__(ar)
            self._x0    = ar.word()
            self._y0    = ar.word()
            self._x1    = ar.word()
            self._y1    = ar.word()
            self._line  = ar.dword()
            self._width = ar.word()

    def svg(self,dwg,g,defmap):
        '''Create SVG object'''
        g.add(dwg.line(start=(self._x0,self._y0),
                       end=(self._x1,self._y1),
                       stroke=self.hex(self._line),
                       stroke_width=self._width))
        
              
    def __str__(self):
        return 'Line: ' + super(GBXLine,self).__str__()
        # f'({self._x0},{self._y0}) -> ({self._x1},{self._y1})')
    
        
# --------------------------------------------------------------------
class GBXTile(GBXDraw):
    def __init__(self,ar):
        '''Draw a tile'''
        with VerboseGuard(f'Reading tile'):
            super(GBXTile,self).__init__(ar)
            self._id = ar.word()

    def svgDef(self,dwg,tileManager,markManager,defmap):
        '''Create SVG definition from image'''
        if self._id in defmap:
            return

        img  = tileManager.image(self._id)
        data = GBXImage.b64encode(img)
        if data is None:
            return

        iden = f'tile_{self._id:04x}'
        img  = dwg.defs.add(dwg.image(id=iden,href=(data),
                                      size=(img.width,img.height)))
        defmap[self._id] = img
        
    def svg(self,dwg,g,defmap):
        '''Create SVG object'''
        if self._id not in defmap: return
        
        g.add(dwg.use(defmap[self._id],
                      insert=(self._left,self._top)))
        
    def __str__(self):
        return f'Tile: {self._id} ' + super(GBXTile,self).__str__()
    
# --------------------------------------------------------------------
class GBXText(GBXDraw):
    def __init__(self,ar):
        '''Draw text'''
        with VerboseGuard(f'Reading text'):
            super(GBXText,self).__init__(ar)
            self._angle   = ar.word()
            self._color   = ar.dword()
            self._text    = ar.str()
            self._font    = CbFont(ar)

    def svg(self,dwg,g,defmap):
        '''Create SVG object'''
        g.add(dwg.text(self._text,
                       insert=(self._left,self._bottom),
                       rotate=[self._angle],
                       fill=self.hex(self._color),
                       font_family='monospace' if self._font._name == '' else self._font._name,
                       font_size=self._font._size,
                       font_weight='bold' if self._font.isBold() else 'normal',
                       font_style='italic' if self._font.isItalic() else 'normal',
                       text_decoration='underline' if self._font.isUnderline() else 'none'))
        
    def __str__(self):
        return f'Text: "{self._text}" '+super(GBXText,self).__str__()
    
# --------------------------------------------------------------------
class GBXPolyline(GBXDraw):
    def __init__(self,ar):
        '''Draw a polyline'''
        with VerboseGuard(f'Reading polyline'):
            super(GBXPolyline,self).__init__(ar)
            self._fill   = ar.dword()
            self._line   = ar.dword()
            self._width  = ar.word()
            n            = (ar.word() if Features().size_size != 8 else
                            ar.size())
            self._points = [[ar.word(),ar.word()] for _ in range(n)]

    def svg(self,dwg,g,defmap):
        '''Create SVG object'''
        g.add(dwg.polyline(self._points,
                           fill=self.hex(self._fill),
                           stroke=self.hex(self._line),
                           stroke_width=self._width))

    def __str__(self):
        return f'Polyline: {len(self._points)} '+super(GBXPolyline,self).__str__()
    
# --------------------------------------------------------------------
class GBXBitmap(GBXDraw):
    CNT = 0
    
    def __init__(self,ar):
        '''Draw a bitmap'''
        with VerboseGuard(f'Reading bitmap'):
            super(GBXBitmap,self).__init__(ar)
            sav = f'B{GBXBitmap.CNT:04d}.png'
            GBXBitmap.CNT += 1
            self._scale = ar.word()
            self._img   = GBXImage(ar,save=None)

    def svg(self,dwg,g,defmap):
        '''Create SVG object'''
        data = GBXImage.b64encode(self._img._img)
        size = self._img._img.width, self._img._img.height
        g.add(dwg.image(insert=(self._left,self._top),
                        size=size,
                        href=(data)))
        
    def __str__(self):
        return f'Bitmap: {self._img} ' + super(GBXBitmap,self).__str__()
    
# --------------------------------------------------------------------
class GBXPiece(GBXDraw):
    def __init__(self,ar):
        '''Draw a piece'''
        with VerboseGuard(f'Reading piece (draw)'):
            super(GBXPiece,self).__init__(ar)
            self._id = ar.iden()

    def toDict(self,calc=None):
        d = {'type': 'Piece',
             'id':    self._id,
             'pixel': self.baseDict()
             }
        if calc is not None:
            d['grid'] = calc(*self.center())
        return d

    def __str__(self):
        return f'Piece: {self._id} ' + super(GBXPiece,self).__str__()
    
# --------------------------------------------------------------------
class GBXMark(GBXDraw):
    def __init__(self,ar):
        '''Draw a mark tile'''
        with VerboseGuard(f'Reading mark (draw)'):
            super(GBXMark,self).__init__(ar)
            self._id  = ar.size()
            self._mid = ar.iden()
            self._ang = ar.word()

    def toDict(self,calc=None):
        d = {'type': 'Mark',
             'id':    self._mid,
             'pixel': self.baseDict()
             }
        if calc is not None:
            d['grid'] = calc(*self.center())
        return d
    def svgDef(self,dwg,tileManager,markManager,defmap):
        '''Create SVG def from mark'''
        if self._id in defmap:
            return

        data = GBXImage.b64encode(tileManager.image(self._id))
        if data is None:
            return

        iden = f'mark_{self._id:04x}'
        img  = dwg.defs.add(dwg.image(id=iden,href=(data)))
        defmap[self._id] = img
        
    def svg(self,dwg,g,defmap):
        '''Create SVG object'''
        if self._id not in defmap: return
        
        g.add(dwg.use(defmap[self._id],
                      insert=(self._left,self._top)))
        
    def __str__(self):
        return f'Mark: {self._id}/{self._mid} ' + super(GBXMark,self).__str__()
    
# --------------------------------------------------------------------
class GBXLineObj(GBXLine):
    def __init__(self,ar):
        '''Line object via reference'''
        with VerboseGuard(f'Reading line object'):
            super(GBXLineObj,self).__init__(ar)

            self._id  = ar.iden()

    def __str__(self):
        return f'Line: {self._id} ' + super(GBXLineObj,self).__str__()

    
# --------------------------------------------------------------------
class GBXDrawList:
    RECT     = 0
    ELLIPSE  = 1
    LINE     = 2
    TILE     = 3
    TEXT     = 4
    POLYLINE = 5
    BITMAP   = 6
    PIECE    = 0x80
    MARK     = 0x81
    LINEOBJ  = 0x82

    TMAP    = { RECT:      GBXRectangle,
                ELLIPSE:   GBXEllipse,
                LINE:      GBXLine,
                TILE:      GBXTile,
                TEXT:      GBXText,
                POLYLINE:  GBXPolyline,
                BITMAP:    GBXBitmap,
                PIECE:     GBXPiece,
                MARK:      GBXMark,
                LINEOBJ:   GBXLineObj}

    def __init__(self,ar):
        '''A list of drawn objects'''
        with VerboseGuard(f'Reading draw list'):
            n = ar.sub_size()

            self._obj = [self._readObj(ar) for n in range(n)]

    def toDict(self,calc=None):
        with VerboseGuard(f'Making dictionary from draw list at pass'):
            ret = []
            for i in self._obj:
                d = i.toDict(calc)
                if d is None:
                    continue

                ret.append(d)

            return ret
        
    def _readObj(self,ar):
        '''Read one object'''
        tpe  = ar.word()
        cls  = self.TMAP.get(tpe,None)
        if cls is None:
            raise RuntimeError(f'Unknown type of draw: {tpe}')

        return cls(ar)

    def svgDefs(self,dwg,tileManager,markManager,defmap):
        '''Create SVG defs'''
        with VerboseGuard(f'Create SVG defs from draw list'):
            for i in self._obj:
                if type(i) not in [GBXTile,GBXMark]: continue

                i.svgDef(dwg,tileManager,markManager,defmap)

    def svg(self,dwg,g,passNo,defmap):
        '''Create SVG objects'''
        with VerboseGuard(f'Drawing SVG from draw list at pass {passNo}'
                          f' ({len(self._obj)} objects)') as gg:
            for i in self._obj:
                if passNo == 1 and i.isSecondPass():
                    continue
                elif passNo == 2 and not i.isSecondPass():
                    continue
                gg(f'Drawing {i}')
                i.svg(dwg,g,defmap)
            
    def __str__(self):
        return '\n        '.join([str(o) for o in self._obj])

#
# EOF
#



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
## END_IMPORT

# --------------------------------------------------------------------
class GlobalCommandTrait(Trait):
    ID = 'globalkey'
    def __init__(self,
                 commandName   = '',
                 key           = '', # Command received
                 globalKey     = '', # Command to send to targets
                 properties    = '', # Filter target on this expression
                 ranged        = False,
                 range         = 1,
                 reportSingle  = True,
                 fixedRange    = True,
                 rangeProperty = '',
                 description   = '',
                 deckSelect    = '-1',
                 target        = ''):
        '''Create a global key command in piece.  This sends a key
        command to near-by counters, as if invoked by a global key
        (module window) command.

        This is _different_ from GlobalHotkeyTrait in that this does
        not invoke and actual module window global hot key, but rather
        sends the command directly to a near-by counter (just how
        close depends on the range or rangeProperty parameter).

        The `deckSelect` select either _all_ (value -1) or specified
        number of pieces from deck.
        
        (VASSAL.counters.CounterGlobalKeyCommand)

        '''
        self.setType(commandName   = commandName,
                     key           = key,
                     globalKey     = globalKey,
                     properties    = properties,
                     ranged        = ranged,
                     range         = range,
                     reportSingle  = reportSingle,
                     fixedRange    = fixedRange,
                     rangeProperty = rangeProperty,
                     description   = description,
                     deckSelect    = deckSelect,
                     target        = target)
        self.setState()
        
Trait.known_traits.append(GlobalCommandTrait)

#
# EOF
#



## BEGIN_IMPORT
from .. common import VerboseGuard
from . base import *
from . element import Element
from . gameelements import GameElement
from . map import WidgetMap
from . withtraits import *
## END_IMPORT

# --------------------------------------------------------------------
class WidgetElement:
    UNIQUE = ['entryName']
    def __init__(self):
        pass

    def addTabs(self,**kwargs):
        '''Add a `Tabs` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Tabs
            The added element
        '''
        return self.add(TabWidget,**kwargs)
    def addCombo(self,**kwargs):
        '''Add a drop-down menu to this

        Parameters
        ----------
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Combo
            The added element
        '''
        return self.add(ComboWidget,**kwargs)
    def addPanel(self,**kwargs):
        '''Add a `Panel` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Panel
            The added element
        '''
        return self.add(PanelWidget,**kwargs)
    def addList(self,**kwargs):
        '''Add a `List` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : List
            The added element
        '''
        return self.add(ListWidget,**kwargs)
    def addMapWidget(self,**kwargs):
        '''Add a `MapWidget` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : MapWidget
            The added element
        '''
        return self.add(MapWidget,**kwargs)
    def addChart(self,**kwargs):
        '''Add a `Chart` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Chart
            The added element
        '''
        ## BEGIN_IMPORT
        from . chart import Chart
        ## END_IMPORT
        return self.add(Chart,**kwargs)
    def addPieceSlot(self,**kwargs):
        '''Add a `PieceSlot` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : PieceSlot
            The added element
        '''
        return self.add(PieceSlot,**kwargs)
    def addPiece(self,piece):
        '''Add a `Piece` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Piece
            The added element
        '''
        if not isinstance(piece,PieceSlot):
            print(f'Trying to add {type(piece)} to ListWidget')
            return None
            
        p = piece.clone(self)
        return p
    def getTabs(self,asdict=True):
        '''Get all Tab element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `Tab` elements.  If `False`, return a list of all Tab` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `Tab` children
        '''
        return self.getElementsByKey(TabWidget,'entryName',asdict)
    def getCombos(self,asdict=True):
        '''Get all Combo element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `Tab` elements.  If `False`, return a list of all Tab` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `Tab` children
        '''
        return self.getElementsByKey(ComboWidget,'entryName',asdict)
    def getLists(self,asdict=True):
        '''Get all List element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `List` elements.  If `False`, return a list of all List` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `List` children
        '''
        return self.getElementsByKey(ListWidget,'entryName',asdict)
    def getPanels(self,asdict=True):
        '''Get all Panel element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `Panel` elements.  If `False`, return a list of all Panel` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `Panel` children
        '''
        return self.getElementsByKey(PanelWidget,'entryName',asdict)
    def getMapWidgets(self,asdict=True):
        '''Get all MapWidget element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `MapWidget` elements.  If `False`, return a list of all MapWidget` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `MapWidget` children
        '''
        return self.getElementsByKey(MapWidget,'entryName',asdict)
    def getCharts(self,asdict=True):
        '''Get all Chart element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `Chart` elements.  If `False`, return a list of all Chart` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `Chart` children
        '''
        return self.getElementsByKey(Chart,'chartName',asdict)
    def getPieceSlots(self,asdict=True):
        '''Get all PieceSlot element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `PieceSlot` elements.  If `False`, return a list of all PieceSlot` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `PieceSlot` children
        '''
        return self.getElementsByKey(PieceSlot,'entryName',asdict)

# --------------------------------------------------------------------
class PieceWindow(GameElement,WidgetElement):
    TAG=Element.MODULE+'PieceWindow'
    UNIQUE = ['name']
    def __init__(self,elem,node=None,
                 name         = '',
                 defaultWidth = 0,
                 hidden       = False,
                 hotkey       = key('C',ALT),
                 scale        = 1.,
                 text         = '',
                 tooltip      = 'Show/hide piece window',
                 icon         = '/images/counter.gif'):
        super(PieceWindow,self).__init__(elem,self.TAG,node=node,
                                         name         = name,
                                         defaultWidth = defaultWidth,
                                         hidden       = hidden,
                                         hotkey       = hotkey,
                                         scale        = scale,
                                         text         = text,
                                         tooltip      = tooltip,
                                         icon         = icon)

registerElement(PieceWindow)

# --------------------------------------------------------------------
class ComboWidget(Element,WidgetElement):
    TAG=Element.WIDGET+'BoxWidget'
    def __init__(self,elem,node=None,entryName='',width=0,height=0):
        super(ComboWidget,self).__init__(elem,
                                       self.TAG,
                                       node = node,
                                       entryName = entryName,
                                       width     = width,
                                       height    = height)
        
registerElement(ComboWidget)

# --------------------------------------------------------------------
class TabWidget(Element,WidgetElement):
    TAG=Element.WIDGET+'TabWidget'
    def __init__(self,elem,node=None,entryName=''):
        super(TabWidget,self).__init__(elem,
                                       self.TAG,
                                       node      = node,
                                       entryName = entryName)

registerElement(TabWidget)

# --------------------------------------------------------------------
class ListWidget(Element,WidgetElement):
    TAG=Element.WIDGET+'ListWidget'
    def __init__(self,elem,node = None,
                 entryName = '',
                 height    = 0,
                 width     = 300,
                 scale     = 1.,
                 divider   = 150):
        super(ListWidget,self).__init__(elem,self.TAG,node=node,
                                        entryName = entryName,
                                        height    = height,
                                        width     = width,
                                        scale     = scale,
                                        divider   = divider)

registerElement(ListWidget)

# --------------------------------------------------------------------
class PanelWidget(Element,WidgetElement):
    TAG=Element.WIDGET+'PanelWidget'
    def __init__(self,elem,node=None,
                 entryName = '',
                 fixed     = False,
                 nColumns  = 1,
                 vert      = False):
        super(PanelWidget,self).__init__(elem,self.TAG,node=node,
                                         entryName = entryName,
                                         fixed     = fixed,
                                         nColumns  = nColumns,
                                         vert      = vert)

registerElement(PanelWidget)

# --------------------------------------------------------------------
class MapWidget(Element):
    TAG=Element.WIDGET+'MapWidget'
    def __init__(self,elem,node=None,entryName=''):
        super(MapWidget,self).__init__(elem,self.TAG,
                                       node      = node,
                                       entryName = entryName)

    def addWidgetMap(self,**kwargs):
        '''Add a `WidgetMap` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : WidgetMap
            The added element
        '''
        return self.add(WidgetMap,**kwargs)
    def getWidgetMaps(self,asdict=True):
        '''Get all WidgetMap element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `WidgetMap` elements.  If `False`, return a list of all WidgetMap` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `WidgetMap` children
        '''
        return self.getElementsByKey(WidgetMap,'mapName',asdict=asdict)
    
registerElement(MapWidget)

#
# EOF
#



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
## END_IMPORT

# --------------------------------------------------------------------
class BasicTrait(Trait):
    ID = 'piece'
    def __init__(self,
                 name      = '',
                 filename  = '',  # Can be empty
                 gpid      = '',  # Can be empty
                 cloneKey  = '',  # Deprecated
                 deleteKey = ''): # Deprecated
        '''Create a basic unit (VASSAL.counters.BasicPiece)'''
        self.setType(cloneKey  = cloneKey,   # CLONEKEY
                     deleteKey = deleteKey,  # DELETEKEY
                     filename  = filename,   # IMAGE  
                     name      = name)       # NAME
        self.setState(map        = 'null', # MAPID (possibly 'null')
                      x          = 0,
                      y          = 0,
                      gpid       = gpid,
                      properties = 0) # PROPERTY COUNT (followed by [; KEY; VALUE]+)

    def getProperties(self):
        n = int(self._state[4])
        return {k: v for k, v in zip(self._state[5::2],
                                     self._state[6::2])}
        
Trait.known_traits.append(BasicTrait)

#
# EOF
#
